#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import json
import os
from copy import copy
from pathlib import Path
import requests

from .interfaces import ModelProvider
from .DynamicDLModel import DynamicDLModel
from typing import IO, Callable, List, Union, Optional
import threading
import time
import datetime
from .misc import calculate_file_hash

UPLOAD_RETRIES = 3
TIME_BETWEEN_RETRIES = 10


def upload_model(url_base, filename, model_name, api_key, dice):
    print('Calculating hash...')
    file_hash = calculate_file_hash(filename)
    print(file_hash)
    for retries in range(UPLOAD_RETRIES):
        print(f"Sending {filename}")
        with open(filename, 'rb') as f:
            files = {'model_binary': f}
            r = requests.post(url_base + "upload_model",
                              files=files,
                              data={"model_type": model_name,
                                    "api_key": api_key,
                                    "dice": dice,
                                    "hash": file_hash})
        print(f"status code: {r.status_code}")
        try:
            print(f"message: {r.json()['message']}")
        except:
            pass

        if r.status_code == 200:
            print("upload successful") # success
            break
        print('Upload error')
        time.sleep(TIME_BETWEEN_RETRIES)
    os.remove(filename)


def upload_data(url_base, filename, api_key):
    for retries in range(UPLOAD_RETRIES):
        print(f"Sending {filename}")
        with open(filename, 'rb') as f:
            files = {'data_binary': f}
            r = requests.post(url_base + "upload_data",
                              files=files,
                              data={"api_key": api_key})
        print(f"status code: {r.status_code}")
        try:
            print(f"message: {r.json()['message']}")
        except:
            pass

        if r.status_code == 200:
            print("upload successful") # success
            break
        print('Upload error')
        time.sleep(TIME_BETWEEN_RETRIES)
    os.remove(filename)


class RemoteModelProvider(ModelProvider):
    
    def __init__(self, models_path, url_base, api_key, temp_upload_dir, delete_old_models = True):
        self.models_path = Path(models_path)
        self.url_base = url_base
        self.api_key = api_key
        self.temp_upload_dir = temp_upload_dir
        self.delete_old_models = delete_old_models
        os.makedirs(self.models_path, exist_ok=True)
        print(f"Config: {self.url_base}, {self.api_key}")

    def load_model(self, model_name: str, progress_callback: Optional[Callable[[int, int], None]] = None,
                   force_download: bool = False,
                   timestamp: Optional[Union[int,str]] = None) -> DynamicDLModel:
        """
        Load latest model from remote server if it does not already exist locally.

        Parameters
        ----------
        model_name : str
            The name of the model to load.
        progress_callback: Callable[[int, int], None] (optional)
            Callback function for progress
        force_download: bool
            Sets the forced redownload of models
        timestamp: int or None
            Return a specific model version (default: latest)

        Returns
        -------
        The model object.
        """
        print(f"Loading model: {model_name}")

        json_content = self.model_details(model_name)

        if json_content is None:
            return None

        # Get the name of the latest model
        r = requests.post(self.url_base + "info_model",
                          json={"model_type": model_name,
                                "api_key": self.api_key})

        if timestamp is None:
            timestamp = json_content['latest_timestamp']

        timestamp = str(timestamp)

        try:
            hash_dict = json_content['hashes']
            file_hash_remote = hash_dict[timestamp]
        except:
            file_hash_remote = json_content['hash'] # fallback to latest hash

        # Check if model already exists locally
        local_model_path = self.models_path / f"{model_name}_{timestamp}.model"
        if os.path.exists(local_model_path) and not force_download:
            print("Model already downloaded. Checking hash...")
            file_hash_local = calculate_file_hash(local_model_path)
            if file_hash_local == file_hash_remote:
                print('Model exists, skipping download')
                model = DynamicDLModel.Load(open(local_model_path, 'rb'))
                return model
            else:
                print('Local model is corrupt')
                os.remove(local_model_path)

        print("Downloading new model...")

        # Receive model
        r = requests.post(self.url_base + "get_model",
                          json={"model_type": model_name,
                                "timestamp": timestamp,
                                "api_key": self.api_key},
                          stream=True)
        success = False
        if r.ok:
            success = True
            total_size_in_bytes = int(r.headers.get('content-length', 0))
            print("Size to download:", total_size_in_bytes)
            block_size = 1024*1024  # 1 MB
            current_size = 0
            with open(local_model_path, 'wb') as file:
                for data in r.iter_content(block_size):
                    current_size += len(data)
                    #print(current_size)
                    if progress_callback is not None:
                        progress_callback(current_size, total_size_in_bytes)
                    file.write(data)

            print("Downloaded size", current_size)
            file_hash_local = calculate_file_hash(local_model_path)

            if current_size != total_size_in_bytes or file_hash_local != file_hash_remote:
                print("Download failed!")
                os.remove(local_model_path)
                success = False

        if success:
            print('Model check OK')
            model = DynamicDLModel.Load(open(local_model_path, "rb"))

            if self.delete_old_models:
                # Deleting older models
                old_models = self.models_path.glob(f"{model_name}_*.model")
                print("Deleting old models: ")
                for old_model in old_models:
                    if old_model != local_model_path:
                        print(f"  Deleting: {str(old_model)}")
                        os.remove(old_model)
            return model
        else:
            print("ERROR: Request to server failed")
            print(f"status code: {r.status_code}")
            try:
                print(f"message: {r.json()['message']}")
            except:
                pass
            return None

    def model_details(self, model_name: str) -> dict:
        # get model versions
        # Get the name of the latest model
        r = requests.post(self.url_base + "info_model",
                          json={"model_type": model_name,
                                "api_key": self.api_key})
        if r.ok:
            json_content = r.json()
        else:
            print("ERROR: Request to server failed")
            print(f"status code: {r.status_code}")
            try:
                print(f"message: {r.json()['message']}")
            except:
                pass
            return None

        # store json data in cache without the model-specific parts
        json_out = copy(json_content)
        keys_to_delete = ['hash', 'hashes', 'latest_timestamp', 'timestamps']
        for k in keys_to_delete:
            try:
                del json_out[k]
            except KeyError:
                pass

        json.dump(json_out, open(self.models_path / f'{model_name}.json', 'w'))

        print("Model details:", json_content)

        return json_content

    def available_models(self) -> Union[None, List[str]]:
        r = requests.post(self.url_base + "get_available_models",
                          json={"api_key": self.api_key})
        if r.ok:
            models = r.json()['models']
            return models
        else:
            print(f"status code: {r.status_code}")
            try:
                print(f"message: {r.json()['message']}")
            except:
                pass
            if r.status_code == 401:
                raise PermissionError("Your api_key is invalid.")
            return None

    def upload_model(self, model_name: str, model: DynamicDLModel, dice_score: float = 0.0):
        """
        Upload model to server
        
        Args:
            model_name: classifier | thigh | leg
            model: DynamicDLModel
        """
        print("Uploading model...")
        filename_out = os.path.join(self.temp_upload_dir, f'{model_name}_{model.timestamp_id}.model')
        model.dump(open(filename_out, 'wb'))
        upload_thread = threading.Thread(target=upload_model, args=(self.url_base, filename_out, model_name,
                                                                    self.api_key, dice_score))
        upload_thread.start()

    def _upload_bytes(self, data: IO):
        # Note: the don't pass data directly to requests because the byte stream is not at the start.
        # Use getbuffer or getvalue instead. See https://github.com/psf/requests/issues/2589
        print("Uploading data")
        filename = datetime.datetime.now().strftime("data_%Y%m%d_%H%M%S.npz")
        filename_out = os.path.join(self.temp_upload_dir, filename)
        with open(filename_out, 'wb') as f:
            f.write(data.getbuffer())
        upload_thread = threading.Thread(target=upload_data, args=(self.url_base, filename_out, self.api_key))
        upload_thread.start()

    def log(self, msg: str):
        r = requests.post(self.url_base + "log",
                        json={"api_key": self.api_key,
                              "message": str(msg)})

        if not r.ok:
            if r.status_code == 401:
                raise PermissionError("Your api_key is invalid.")
            else:
                raise OSError("Error communicating with server")




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Implementation of a deep learning module that can be serialized and deserialized, and dynamically changed.
Functions for the operation of the class are provided as references to top-level functions.
Such top level functions should define all the imports within themselves (i.e. don't put the imports at the top of the file).
"""
#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import re

from .interfaces import IncompatibleModelError, DeepLearningClass
import dill
from io import BytesIO
import numpy as np
import inspect
import time

def fn_to_source(function):
    """
    Given a function, returns it source. If the source cannot be retrieved, return the object itself
    """
    #print('Converting fn to source')
    if function is None: return None
    try:
        return inspect.getsource(function)
    except OSError:
        #print('Conversion failed!')
        try:
            src = function.source
            #print('Returning embedded source')
            return src
        except:
            pass
    print('Getting source failed - Returning bytecode')
    return function # the source cannot be retrieved, return the object itself


def source_to_fn(source, patches: dict = {}):
    """
    Given a source, return the (first) defined function. If the source is not a string, return the object itself
    """
    if type(source) is not str:
        print("source to fn: source is not a string")
        return source
    #print("source to fn: source is string")
    for search, replace in patches.items():
        source = re.sub(search, replace, source)

    locs = {}
    globs = {}
    try:
        exec(source, globs, locs)
    except Exception as e:
        print("source to fn: exec failed", e)
        return source # the string was just a string apparently, not valid code
    for k,v in locs.items():
        if callable(v):
            #print('source_to_fn. Found function', k)
            v.source = source
            return v
    return source


def default_keras_weights_to_model_function(modelObj: DynamicDLModel, weights):
    modelObj.model.set_weights(weights)


def default_keras_model_to_weights_function(modelObj: DynamicDLModel):
    return modelObj.model.get_weights()


def default_keras_delta_function(lhs: DynamicDLModel, rhs: DynamicDLModel, threshold=None):
    from dafne_dl.interfaces import IncompatibleModelError
    if lhs.model_id != rhs.model_id: raise IncompatibleModelError
    lhs_weights = lhs.get_weights()
    rhs_weights = rhs.get_weights()
    newWeights = []
    for depth in range(len(lhs_weights)):
        delta = lhs_weights[depth] - rhs_weights[depth]
        if threshold is not None:
            delta[np.abs(delta) < threshold] = 0
        newWeights.append(delta)
    outputObj = lhs.get_empty_copy()
    outputObj.set_weights(newWeights)
    outputObj.is_delta = True
    outputObj.timestamp_id = rhs.timestamp_id # set the timestamp of the original model to identify the base
    return outputObj


def default_keras_add_weights_function(lhs: DynamicDLModel, rhs: DynamicDLModel):
    from dafne_dl.interfaces import IncompatibleModelError
    if lhs.model_id != rhs.model_id: raise IncompatibleModelError
    lhs_weights = lhs.get_weights()
    rhs_weights = rhs.get_weights()
    newWeights = []
    for depth in range(len(lhs_weights)):
        newWeights.append(lhs_weights[depth] + rhs_weights[depth])
    outputObj = lhs.get_empty_copy()
    outputObj.set_weights(newWeights)
    return outputObj


def default_keras_multiply_function(lhs: DynamicDLModel, rhs: float):
    if not isinstance(rhs, (int, float)):
        raise NotImplementedError('Incompatible types for multiplication (only multiplication by numeric factor is allowed)')
    lhs_weights = lhs.get_weights()
    newWeights = []
    for depth in range(len(lhs_weights)):
        newWeights.append(lhs_weights[depth] * rhs)
    outputObj = lhs.get_empty_copy()
    outputObj.set_weights(newWeights)
    return outputObj


def default_keras_weight_copy_function(weights_in):
    weights_out = []
    for layer in weights_in:
        weights_out.append(layer.copy())
    return weights_out


class DynamicDLModel(DeepLearningClass):

    """
    Class to represent a deep learning model that can be serialized/deserialized
    """
    def __init__(self, model_id,  # a unique ID to avoid mixing different models
                 init_model_function,  # inits the model. Accepts no parameters and returns the model
                 apply_model_function,  # function that applies the model. Has the object, and image
                 weights_to_model_function = default_keras_weights_to_model_function,  # put model weights inside the model.
                 model_to_weights_function = default_keras_model_to_weights_function,  # get the weights from the model in a pickable format
                 calc_delta_function = default_keras_delta_function,  # calculate the weight delta
                 apply_delta_function = default_keras_add_weights_function,  # apply a weight delta
                 weight_copy_function = default_keras_weight_copy_function,  # create a deep copy of weights
                 factor_multiply_function = default_keras_multiply_function,
                 incremental_learn_function = None,  # function to perform an incremental learning step
                 weights = None,  # initial weights
                 timestamp_id = None,
                 is_delta = False):
        self.model = None
        self.model_id = model_id
        self.is_delta = is_delta

        # lsit identifying the external functions that need to be saved with source and serialized
        self.function_mappings = [
            'init_model_function',
            'apply_model_function',
            'weights_to_model_function',
            'model_to_weights_function',
            'calc_delta_function',
            'apply_delta_function',
            'weight_copy_function',
            'factor_multiply_function',
            'incremental_learn_function',
        ]

        # the following sets the internal attributes self.fn = fn, with additionally adding the source to the function
        for fn_name in self.function_mappings:
            self.set_internal_fn(fn_name, locals()[fn_name])


        self.init_model() # initializes the model
        if timestamp_id is None:
            self.reset_timestamp()
        else:
            self.timestamp_id = timestamp_id  # unique timestamp id; used to identify model versions during federated learning

        if weights: self.set_weights(weights)

    def set_internal_fn(self, internal_name, obj):
        #print('Setting', internal_name)
        if callable(obj):
            src = fn_to_source(obj)
            if type(src) == str:
                obj.source = src

        setattr(self, internal_name, obj)

    def init_model(self):
        """
        Initializes the internal model

        Returns
        -------
        None.

        """
        self.model = self.init_model_function()
        
    def set_weights(self, weights):
        """
        Loads the weights in the internal model

        Parameters
        ----------
        weights : whatever is accepted by the model_to_weights_function
            Weights to be loaded into the model

        Returns
        -------
        None.

        """
        self.weights_to_model_function(self, weights)
        
    def get_weights(self):
        return self.model_to_weights_function(self)
        
    def apply_delta(self, other):
        return self.apply_delta_function(self, other)
    
    def calc_delta(self, other, threshold=None):
        return self.calc_delta_function(self, other, threshold)
    
    def apply(self, data):
        return self.apply_model_function(self, data)

    def factor_multiply(self, factor: float):
        return self.factor_multiply_function(self, factor)
    
    def incremental_learn(self, trainingData, trainingOutputs, bs=5, minTrainImages=5):
        self.incremental_learn_function(self, trainingData, trainingOutputs, bs, minTrainImages)
        
    def dump(self, file):
        """
        Dumps the current status of the object, including functions and weights
        
        Parameters
        ----------
        file:
            a file descriptor (open in writable mode)

        Returns
        -------
        Nothing

        """
        outputDict = {
            'model_id': self.model_id,
            'weights': self.get_weights(),
            'timestamp_id': self.timestamp_id,
            'is_delta': self.is_delta
            }

        # add the internal functions to the dictionary
        for fn_name in self.function_mappings:
            outputDict[fn_name] = fn_to_source(getattr(self, fn_name))

        dill.dump(outputDict, file)
    
    def dumps(self) -> bytes:
        file = BytesIO()
        self.dump(file)
        return file.getvalue()
    
    def get_empty_copy(self) -> DynamicDLModel:
        """
        Gets an empty copy (i.e. without weights) of the current object

        Returns
        -------
        DynamicDLModel
            Output copy

        """
        new_model = DynamicDLModel(self.model_id, self.init_model_function, self.apply_model_function,
                                   weights=None, timestamp_id=self.timestamp_id, is_delta=self.is_delta)
        for fn_name in self.function_mappings:
            new_model.set_internal_fn(fn_name, getattr(self, fn_name))
        return new_model

    def copy(self) -> DynamicDLModel:
        """
        Gets a copy (i.e. with weights) of the current object

        Returns
        -------
        DynamicDLModel
            Output copy

        """
        model_out = self.get_empty_copy()
        model_out.set_weights( self.weight_copy_function(self.get_weights()) )
        return model_out

    @staticmethod
    def Load(file) -> DynamicDLModel:
        """
        Creates an object from a file

        Parameters
        ----------
        file : file descriptor
            A file descriptor.

        Returns
        -------
        DynamicDLModel
            A new instance of a dynamic model

        """

        # code patches for on-the-fly conversion of old models to new format
        patches = {
            'from dl': 'from dafne_dl',
            'import dl': 'import dafne_dl'
        }

        inputDict = dill.load(file)
        for k,v in inputDict.items():
            if '_function' in k:
                inputDict[k] = source_to_fn(v, patches) # convert the functions from source

        #print(inputDict)
        outputObj = DynamicDLModel(**inputDict)
        return outputObj
        
    @staticmethod
    def Loads(b: bytes) -> DynamicDLModel:
        """
        Creates an object from a binary dump

        Parameters
        ----------
        file : bytes
            A sequence of bytes

        Returns
        -------
        DynamicDLModel
            A new instance of a dynamic model

        """
        file = BytesIO(b)
        return DynamicDLModel.Load(file)



#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .utils import invert_dict, merge_dict

short_labels = {
    1: 'VL',
    2: 'VM',
    3: 'VI',
    4: 'RF',
    5: 'SAR',
    6: 'GRA',
    7: 'AM',
    8: 'SM',
    9: 'ST',
    10: 'BFL',
    11: 'BFS',
    12: 'AL'
}

long_labels = {
    1: 'Vastus Lateralis',
    2: 'Vastus Medialis',
    3: 'Vastus Intermedius',
    4: 'Rectus Femoris',
    5: 'Sartorius',
    6: 'Gracilis',
    7: 'Adductor Magnus',
    8: 'Semimembranosus',
    9: 'Semitendinosus',
    10: 'Biceps Femoris Long',
    11: 'Biceps Femoris Short',
    12: 'Adductor Longus'
}

long_labels_split = {}
ctr = 1
for key, val in long_labels.items():
    long_labels_split[ctr] = val + "_R"
    long_labels_split[ctr+1] = val + "_L"
    ctr += 2

inverse_labels = merge_dict(invert_dict(short_labels), invert_dict(long_labels))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import json
import re
from pathlib import Path

from .interfaces import ModelProvider
from .DynamicDLModel import DynamicDLModel
from typing import Union, IO, List, Optional
import os
import datetime
from typing import Callable


class LocalModelProvider(ModelProvider):

    def __init__(self, models_path, upload_dir):
        self.models_path = Path(models_path)
        self.upload_dir = upload_dir

    def get_model_names(self):
        model_list = self.models_path.glob('*.model')
        model_names = list(set([os.path.basename(s).split('_')[0] for s in
                                model_list]))  # get the name of the model, which is the part of the file before the '_'
        model_names = list(filter(None, model_names))  # remove any empty names
        return model_names

    def load_model(self, model_name: str, progress_callback: Optional[Callable[[int, int], None]] = None,
                   force_download: bool = False,
                   timestamp: Optional[Union[int,str]] = None) -> DynamicDLModel:
        """
        Loads a deep learning model.

        Parameters
        ----------
        model_name : str
            The name of the model to load.
        progress_callback: Callable[[int, int], None] (optional)
            Callback function for progress
        force_download: bool
            Sets the forced redownload of models
        timestamp: int or None
            Return a specific model version (default: latest)

        Returns
        -------
        The model object.

        """
        print(f"Loading model: {model_name}")
        model_file = sorted(list(self.models_path.glob(f"{model_name}_*.model")))
        if len(model_file) == 0:
            raise FileNotFoundError("Could not find model file.")

        model_to_load = None
        if timestamp is None:
            model_to_load = model_file[-1]
        else:
            for filename in model_file:
                if str(timestamp) in filename:
                    model_to_load = filename
                    break

        if model_to_load is None:
            raise FileNotFoundError("Could not find model file.")

        print('Opening', model_to_load)
        return DynamicDLModel.Load(open(model_to_load, 'rb'))

    def model_details(self, model_name: str) -> dict:
        # get model versions
        model_files = sorted(list(self.models_path.glob(f"{model_name}_*.model")))
        timestamps = []
        for filename in model_files:
            try:
                timestamps.append(re.match(r".*([0-9]+)\.model", filename).group(1))
            except:
                pass

        out_dict = {}
        json_file_name = f'{model_name}.json'
        try:
            # add the content of the json file to the dictionary
            out_dict.update( json.load(open(self.models_path / json_file_name, 'rb')) )
        except:
            pass

        out_dict['timestamps'] = timestamps

        return out_dict

    def import_model(self, file_path, model_name):
        print('Importing model')
        model = DynamicDLModel.Load(open(file_path, 'rb'))
        self.upload_model(model_name, model)

    def available_models(self) -> Union[None, List[str]]:
        return self.get_model_names()

    def upload_model(self, model_name: str, model: DynamicDLModel, dice_score: float = 0.0):
        print("You are using the LocalModelProvider. Model is saved in the model directory!")
        filename = f'{model_name}_{model.timestamp_id}.model'
        print('Saving', filename)
        model.dump(open(os.path.join(self.models_path, filename), 'wb'))

    def _upload_bytes(self, data: IO):
        print("You are using the LocalModelProvider. Therefore no upload is done!")
        filename = datetime.datetime.now().strftime("data_%Y%m%d_%H%M%S.npz")
        with open(os.path.join(self.upload_dir, filename), 'wb') as f:
            f.write(data.getbuffer())
        print('File saved')

    def log(self, msg: str):
        print("LocalModelProvider log", msg)



#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Data generator classes - from directory or from memory
#from collections import Sequence
from tensorflow.keras.utils import Sequence
import numpy as np
import os

# This is NOT a general implementation! See the number of masks below
class DataGeneratorDir(Sequence):
    def __init__(self, path, list_X=list(range(1, 4501)), batch_size=20, dim=(432, 432), shuffle=True):
        'Initialization'
        self.dim = dim
        self.batch_size = batch_size
        self.list_X = list_X
        self.path = path
        self.shuffle = shuffle
        self.on_epoch_end()

    def __len__(self):
        'Denotes the number of batches per epoch'
        return int(np.floor(len(self.list_X) / self.batch_size))

    def __getitem__(self, index):
        'Generate one batch of data'
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size:(index + 1) * self.batch_size]

        # Find list of IDs
        list_X_temp = [self.list_X[k] for k in indexes]

        # Generate data
        X, y = self.__data_generation(list_X_temp, self.path)

        return X, y

    def on_epoch_end(self):
        'Updates indexes after each epoch'
        self.indexes = np.arange(len(self.list_X))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __data_generation(self, list_X_temp, path):
        'Generates data containing batch_size samples'
        # Initialization
        X = np.empty((self.batch_size, *self.dim, 2))
        y = np.empty((self.batch_size, *self.dim, 13)) # chenge this to be generic!

        # Generate data
        for i, j in enumerate(list_X_temp):
            # Store sample
            arr = np.load(os.path.join(path, 'train_' + str(j) + '.npy'))

            X[i,] = np.stack([arr[:, :, 0], arr[:, :, -1]], axis=-1)

            # Store class
            y[i,] = arr[:, :, 1:-1]

        return X, y

class DataGeneratorMem(Sequence):
    def __init__(self, training_data_list, list_X=list(range(1, 4501)), batch_size=20, dim=(432, 432), shuffle=True):
        #print('Data Generator Initialization. Data list len:', len(training_data_list))
        #print('list_X', list_X)
        self.dim = dim
        self.batch_size = batch_size
        self.list_X = list_X
        self.training_data_list = training_data_list
        self.shuffle = shuffle
        self.on_epoch_end()
        self.n_labels = training_data_list[0].shape[2]-2

    def __len__(self):
        'Denotes the number of batches per epoch'
        return int(np.floor(len(self.list_X) / self.batch_size))

    def __getitem__(self, index):
        'Generate one batch of data'
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size:(index + 1) * self.batch_size]

        # Find list of IDs
        list_X_temp = [self.list_X[k] for k in indexes]
        #print("list_X_temp", list_X_temp)

        # Generate data
        X, y = self.__data_generation(list_X_temp)

        return X, y

    def on_epoch_end(self):
        'Updates indexes after each epoch'
        self.indexes = np.arange(len(self.list_X))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __data_generation(self, list_X_temp):
        'Generates data containing batch_size samples'
        # Initialization
        X = np.empty((self.batch_size, *self.dim, 2))
        y = np.empty((self.batch_size, *self.dim, self.n_labels))

        # Generate data
        for i, j in enumerate(list_X_temp):
            # Store sample
            arr = self.training_data_list[j]

            #X[i,] = np.stack([arr[:, :, 0], arr[:, :, -1]], axis=-1)
            X[i, :, :, 0] = arr[:, :, 0]
            X[i, :, :, 1] = arr[:, :, -1]

            # Store class
            y[i,] = arr[:, :, 1:-1]

        return X, y



# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import os
import skimage
from skimage.morphology import square
from .padorcut import padorcut
from . import biascorrection
from scipy.ndimage import zoom
import math
from skimage.filters import threshold_otsu
import tensorflow as tf
from tensorflow.keras import backend as K


#convert an array of masks into a single numbered mask
def to_mask(categorical_mask,dim,ch):  ##ch: number of labels. ch = 13 for thigh, 7 for leg
    segmentation_mask=np.zeros((dim,dim))
    for i in range(categorical_mask.shape[1]):
        for j in range(categorical_mask.shape[1]):
            segmentation_mask[i,j]=categorical_mask[i,j,1:ch+1].argmax()
    return segmentation_mask


def split_mirror(image):
    '''
    Function to split the training data into left and mirrored right parts cropped and zoomed/padorcutted. 
    '''
    import skimage
    from skimage.morphology import square
    from skimage.filters import threshold_local
    block_size = 15
    # convert image to binary mask
    local_thresh = threshold_local(image, block_size, offset=10)
    binary_local = image > local_thresh
    binary_local=binary_local==0.0
    # remove small black and white spots
    binary_local=skimage.morphology.area_opening(binary_local,area_threshold=4) #20
    binary_local=skimage.morphology.area_closing(binary_local,area_threshold=4) #20
    mountainsc=binary_local[:,:].sum(axis=0)
    mountainsr=binary_local[:,:].sum(axis=1)
    s=0
    p=0
    ii=0 
    jj=0
    a1=0
    a2=0
    a3=0
    a4=0
    b1=0
    b2=0
    #iterate through columns and find transition (black)-white-black-white-black
    while s!=4 and ii<432:
        if mountainsc[ii]>0 and s==0:
            s=1
            a1=ii
        if mountainsc[ii]==0 and s==1 and (ii-a1)>20:
            s=2
            a2=ii
        if mountainsc[ii]>0 and s==2:
            s=3
            a3=ii
        if mountainsc[ii]==0 and s==3 and (ii-a3)>20:
            s=4
            a4=ii
        elif mountainsc[ii]==0 and s==3 and (ii-a3)<20:
            s=2
        ii+=1
        # this means that there is only one patch (legs as attached). Set the splitting point in the middle
        if ii==432 and s==2:
            s=4
            a4=a2
            a2=np.ceil((a4-a1)/2)+a1
            a3=np.ceil((a4-a1)/2)+a1
        if ii==432 and s==3:
            s=4
            a4=431
            #a2=np.ceil((a4-a1)/2)+a1
            #a3=np.ceil((a4-a1)/2)+a1
        if ii==432 and s==1:
            s=4
            a4=431
            a2=np.ceil((a4-a1)/2)+a1
            a3=np.ceil((a4-a1)/2)+a1
    # iterate over columns and find the (black)-white-black transition
    while p!=2:
        if mountainsr[jj]>0 and p==0:
            p=1
            b1=jj
        if mountainsr[jj]==0 and p==1 and (jj-b1)>50:
            p=2
            b2=jj
        jj+=1
    return a1,a2,a3,a4,b1,b2

# returns arrays with number of pixels corresponding to a class and total number of pixels of the images containing that class
def compute_class_frequencies(path,dim,card,ch):
    classes=[0]*ch
    images_containing_class=[0]*ch
    for j in range(1,card+1):
        arr=np.load(os.path.join(path,'train_'+str(j)+'.npy'))
        if arr.shape[2]==2:
            seg=arr[:,:,1] 
        else:
            seg=to_mask(arr,dim,ch)
        for class_ in range(ch):
            n_pixels=(seg==float(class_)).sum()
            if n_pixels>0:
                classes[class_]+=n_pixels
                images_containing_class[class_]+=dim**2
    return classes,images_containing_class


def calc_weight(img, seg, av, freq, dim, band, ch):
    W = np.zeros((seg.shape[0], seg.shape[1]), dtype='float32')
    global_thresh = threshold_otsu(img)
    binary_global = img > global_thresh  # image to binary mask
    binary_mask = binary_global == 1.0
    maximum_l = min(3 * np.sqrt(band), seg.shape[0] - 1)
    for i in range(seg.shape[0]):
        for j in range(seg.shape[1]):
            if (binary_mask[i, j] > 0):
                h = 0
                k = 0
                d1 = 0
                d2 = 0
                aa = seg[i, j]
                l = 1
                while h != 1 or k != 1:
                    # find a border with another patch
                    if h == 0:  # point at [i+-l,j] or [i,j+-l] is neither 0 nor aa
                        if (seg[max(i - l, 0), j] != aa and int(seg[max(i - l, 0), j] != 0)):
                            d1 = l
                            h = 1
                            bb = seg[i - l, j]
                        if (seg[min(i + l, seg.shape[0] - 1), j] != aa and int(
                                seg[min(i + l, seg.shape[0] - 1), j]) != 0):
                            d1 = l
                            h = 1
                            bb = seg[min(i + l, seg.shape[0] - 1), j]
                        if (seg[i, max(j - l, 0)] != aa and int(seg[i, max(j - l, 0)]) != 0):
                            d1 = l
                            h = 1
                            bb = seg[i, j - l]
                        if (seg[i, min(j + l, seg.shape[0] - 1)] != aa and int(
                                seg[i, min(j + l, seg.shape[0] - 1)]) != 0):
                            d1 = l
                            h = 1
                            bb = seg[i, min(j + l, seg.shape[0] - 1)]
                    if h == 1 and k == 0 and (
                    # points at [i+-l, j] and [i, j+-l] are neither aa nor bb nor 0. Distance from another patch?
                            (seg[max(i - l, 0), j] != aa and seg[max(i - l, 0), j] != bb and int(
                                seg[max(i - l, 0), j] != 0)) or
                            (seg[min(i + l, seg.shape[0] - 1), j] != aa and seg[
                                min(i + l, seg.shape[0] - 1), j] != bb and int(
                                seg[min(i + l, seg.shape[0] - 1), j]) != 0) or
                            (seg[i, max(j - l, 0)] != aa and seg[i, max(j - l, 0)] != bb and int(
                                seg[i, max(j - l, 0)]) != 0) or
                            (seg[i, min(j + l, seg.shape[0] - 1)] != aa and seg[
                                i, min(j + l, seg.shape[0] - 1)] != bb and int(
                                seg[i, min(j + l, seg.shape[0] - 1)]) != 0)
                    ):
                        d2 = l
                        k = 1
                    if l >= maximum_l:  # l==seg.shape[0]-1:
                        d1 = seg.shape[0]
                        h = 1
                        k = 1
                    l += 1
                W[i, j] = 10 * math.exp(-((d1 + d2) ** 2) / (2 * band))
            W[i, j] = W[i, j] + av / freq[int(seg[i, j])] 
    return W


def input_creation(path,card,dim,band,ch):
    '''
    Creates the training data with labels categorization and creation of weights maps.
    '''
    classes,images=compute_class_frequencies(path,dim,card,ch)
    frequencies=[]
    for cla,ima in zip(classes,images):
        frequencies.append(cla/ima)
    av=sum(frequencies)/ch
    for j in range(1,card+1):
        arr=np.load(os.path.join(path,'train_'+str(j)+'.npy'))
        img=arr[:,:,0]
        if arr.shape[2]==2:
           seg=arr[:,:,1] 
        else:
           seg=to_mask(arr,dim,ch)
        categ=categorical_and_weight(img,seg,av,frequencies,dim,band,ch)
        arr=np.concatenate([np.reshape(img,(dim,dim,1)),categ],axis=-1)
        np.save(os.path.join(path,'train_'+str(j)+'.npy'),arr)


def calc_aggregated_masks_and_remove_overlap(masks):
    dim = masks.shape[0]
    ch = masks.shape[2]
    # remove overlaps between masks
    cumulated_mask = masks[:, :, 0]
    aggregated_mask = np.zeros_like(cumulated_mask)  # this mask has values from 0 to ch
    for mask_dim in range(1, ch):
        masks[:, :, mask_dim] = np.logical_and(masks[:, :, mask_dim], np.logical_not(cumulated_mask))
        aggregated_mask += mask_dim * masks[:, :, mask_dim]
        cumulated_mask = np.logical_or(cumulated_mask, masks[:, :, mask_dim])

    # set the background mask to the not of the other masks
    masks[:,:,0] = np.logical_not(cumulated_mask)

    return aggregated_mask, masks

def input_creation_mem(image_list: list, mask_list: list, band: float):
    """
    Creates the training data in memory

    Inputs:
        image_list: list of 2D slices (input data)
        mask_list: list of 3D np arrays (stacks of segmented 2D masks). All the masks must have the same number of layers (ROIs)
        band: scalar parameter for the calculation of the weights

    Output:
         list of 3D arrays where: arr[:,:,0] are the base images, arr[:,:,-1] are the weights, and the dimensions in between are the masks
    """

    dim = mask_list[0].shape[0]
    ch = mask_list[0].shape[2]

    aggregated_masks = []
    mask_list_no_overlap = []
    for masks in mask_list:
        agg, new_masks = calc_aggregated_masks_and_remove_overlap(masks)
        aggregated_masks.append(agg)
        mask_list_no_overlap.append(new_masks)

    # compute class frequencies
    classes = [0] * ch
    images_containing_class = [0] * ch
    for seg in aggregated_masks:
        for class_ in range(ch):
            n_pixels = (seg == float(class_)).sum()
            if n_pixels > 0:
                classes[class_] += n_pixels
                images_containing_class[class_] += dim ** 2

    frequencies = []
    for cla, ima in zip(classes, images_containing_class):
        if cla == 0:
            frequencies.append(0)
        else:
            frequencies.append(cla / ima)

    #print("Frequencies", frequencies)

    av = sum(frequencies) / ch

    output_data = []

    for slice_number in range(len(image_list)):
        img = image_list[slice_number]
        seg = aggregated_masks[slice_number]
        weight = calc_weight(img,seg,av,frequencies,dim,band,ch)
        arr=np.concatenate([np.reshape(img,(dim,dim,1)),mask_list_no_overlap[slice_number], np.reshape(weight, (dim,dim,1))], axis=-1)
        output_data.append(arr)

    return output_data

def common_input_process(inverse_label_dict, MODEL_RESOLUTION, MODEL_SIZE, trainingData, trainingOutputs):
    # inverse_label_dict = {v: k for k, v in LABELS_DICT.items()}
    nlabels = len(set(inverse_label_dict.values()))+1 # get the number of unique values in the inverse dict
    min_defined_rois = nlabels/2 # do not add to the training set if less than this number of ROIs are defined
    resolution = np.array(trainingData['resolution'])
    zoomFactor = resolution / MODEL_RESOLUTION

    image_list = []
    mask_list = []
    for imageIndex in range(len(trainingData['image_list'])):
        mask_dataset = np.zeros((MODEL_SIZE[0], MODEL_SIZE[1], nlabels))
        defined_rois = 0
        for label, mask in trainingOutputs[imageIndex].items():
            xl=label.endswith('_L')
            xr=label.endswith('_R')
            base_label = label[:-2]
            if xr:
                if (base_label + '_L') in trainingOutputs[imageIndex]:
                    continue # if laterality is split, merge the right into the left
                label = base_label
            if xl:
                if (base_label + '_R') in trainingOutputs[imageIndex]:
                    mask=np.logical_or(trainingOutputs[imageIndex][base_label +'_L'], trainingOutputs[imageIndex][base_label +'_R'])
                label = base_label

            if label not in inverse_label_dict: continue

            if np.sum(mask) > 5:
                defined_rois += 1
            else:
                continue # avoid adding empty masks

            mask = skimage.morphology.area_opening(mask, area_threshold=4)
            mask = skimage.morphology.area_closing(mask, area_threshold=4)
            mask_dataset[:, :, int(inverse_label_dict[label])] = padorcut(zoom(mask, zoomFactor, order=0), MODEL_SIZE)

        if defined_rois > min_defined_rois:
            mask_list.append(mask_dataset)
            image = trainingData['image_list'][imageIndex]
            image = skimage.morphology.area_opening(image, area_threshold=4)
            image = skimage.morphology.area_closing(image, area_threshold=4)
            image_list.append(padorcut(zoom(image, zoomFactor), MODEL_SIZE))

    return image_list, mask_list

def common_input_process_split(inverse_label_dict, MODEL_RESOLUTION, MODEL_SIZE, MODEL_SIZE_SPLIT, trainingData, trainingOutputs):
    nlabels = len(set(inverse_label_dict.values()))+1 # get the number of unique values in the inverse dict
    min_defined_rois = nlabels/2 # do not add to the training set if less than this number of ROIs are defined
    resolution = np.array(trainingData['resolution'])
    zoomFactor = resolution / MODEL_RESOLUTION

    image_list = []
    mask_list = []
    for imageIndex in range(len(trainingData['image_list'])):
        mask_dataset_left = np.zeros((MODEL_SIZE_SPLIT[0], MODEL_SIZE_SPLIT[1], nlabels))
        mask_dataset_right = np.zeros((MODEL_SIZE_SPLIT[0], MODEL_SIZE_SPLIT[1], nlabels))
        defined_rois = 0

        # first, count the defined ROIS, before doing lengthy calculations
        for label, mask in trainingOutputs[imageIndex].items():
            xl = label.endswith('_L')
            xr = label.endswith('_R')
            base_label = label[:-2]
            if xr:
                if (base_label + '_L') in trainingOutputs[imageIndex]:
                    continue  # if laterality is split, merge the right into the left
                label = base_label
            if xl:
                if (base_label + '_R') in trainingOutputs[imageIndex]:
                    mask = np.logical_or(trainingOutputs[imageIndex][base_label + '_L'],
                                         trainingOutputs[imageIndex][base_label + '_R'])
                label = base_label

            if label not in inverse_label_dict: continue

            if np.sum(mask) > 5:
                defined_rois += 1
            else:
                continue # avoid adding empty masks
           
        if defined_rois > min_defined_rois:
            image = trainingData['image_list'][imageIndex]
            image = skimage.morphology.area_opening(image, area_threshold=4)
            image = skimage.morphology.area_closing(image, area_threshold=4)
            image=padorcut(zoom(image, zoomFactor), MODEL_SIZE)
            imgbc=biascorrection.biascorrection_image(image)
            a1,a2,a3,a4,b1,b2=split_mirror(imgbc)
            left=imgbc[int(b1):int(b2),int(a1):int(a2)]
            left=padorcut(left, MODEL_SIZE_SPLIT)
            right=imgbc[int(b1):int(b2),int(a3):int(a4)]
            right=right[::1,::-1]
            right=padorcut(right, MODEL_SIZE_SPLIT)

            for label, mask in trainingOutputs[imageIndex].items():
                xl = label.endswith('_L')
                xr = label.endswith('_R')
                base_label = label[:-2]
                if xr:
                    if (base_label + '_L') in trainingOutputs[imageIndex]:
                        continue  # if laterality is split, merge the right into the left
                    label = base_label
                if xl:
                    if (base_label + '_R') in trainingOutputs[imageIndex]:
                        mask = np.logical_or(trainingOutputs[imageIndex][base_label + '_L'],
                                             trainingOutputs[imageIndex][base_label + '_R'])
                    label = base_label

                if label not in inverse_label_dict: continue

                mask = skimage.morphology.area_opening(mask, area_threshold=4)
                mask = skimage.morphology.area_closing(mask, area_threshold=4)
                mask = padorcut(zoom(mask, zoomFactor, order=0), MODEL_SIZE)
                roileft=mask[int(b1):int(b2),int(a1):int(a2)]
                roileft=padorcut(roileft, MODEL_SIZE_SPLIT)
                roiright=mask[int(b1):int(b2),int(a3):int(a4)]
                roiright=roiright[::1,::-1]
                roiright=padorcut(roiright, MODEL_SIZE_SPLIT)

                mask_dataset_left[:, :, int(inverse_label_dict[label])] = roileft
                mask_dataset_right[:, :, int(inverse_label_dict[label])] = roiright

            image_list.append(left)
            mask_list.append(mask_dataset_left)
            image_list.append(right)
            mask_list.append(mask_dataset_right)

    return image_list, mask_list


def common_input_process_single(inverse_label_dict, MODEL_RESOLUTION, MODEL_SIZE, MODEL_SIZE_SPLIT, trainingData,
                               trainingOutputs, swap):
    nlabels = len(set(inverse_label_dict.values())) + 1  # get the number of unique values in the inverse dict
    min_defined_rois = nlabels / 2  # do not add to the training set if less than this number of ROIs are defined
    resolution = np.array(trainingData['resolution'])
    zoomFactor = resolution / MODEL_RESOLUTION

    image_list = []
    mask_list = []
    for imageIndex in range(len(trainingData['image_list'])):
        mask_dataset = np.zeros((MODEL_SIZE_SPLIT[0], MODEL_SIZE_SPLIT[1], nlabels))
        defined_rois = 0

        # first, count the defined ROIS, before doing lengthy calculations
        for label, mask in trainingOutputs[imageIndex].items():
            xl = label.endswith('_L')
            xr = label.endswith('_R')
            if xl or xr:
                base_label = label[:-2]
            else:
                base_label = label

            if base_label not in inverse_label_dict: continue

            if np.sum(mask) > 5:
                defined_rois += 1
                #print(base_label, 'Found, total rois:', defined_rois)
            else:
                continue  # avoid adding empty masks

        if defined_rois > min_defined_rois:
            image = trainingData['image_list'][imageIndex]
            image = skimage.morphology.area_opening(image, area_threshold=4)
            image = skimage.morphology.area_closing(image, area_threshold=4)
            image = padorcut(zoom(image, zoomFactor), MODEL_SIZE_SPLIT)
            imgbc = biascorrection.biascorrection_image(image)

            if swap:
                imgbc = imgbc[::1,::-1]

            for label, mask in trainingOutputs[imageIndex].items():
                xl = label.endswith('_L')
                xr = label.endswith('_R')
                if xl or xr:
                    base_label = label[:-2]
                else:
                    base_label = label

                if base_label not in inverse_label_dict: continue

                mask = skimage.morphology.area_opening(mask, area_threshold=4)
                mask = skimage.morphology.area_closing(mask, area_threshold=4)
                mask = padorcut(zoom(mask, zoomFactor, order=0), MODEL_SIZE_SPLIT)

                if swap:
                    mask = mask[::1,::-1]

                mask_dataset[:, :, int(inverse_label_dict[base_label])] = mask

            image_list.append(imgbc)
            mask_list.append(mask_dataset)

    return image_list, mask_list


def weighted_loss(y_true,y_pred):
    weight_matrix=K.flatten(y_pred[:,:,:,-1])
    y_pre=y_pred[:,:,:,:-1]
    E=-K.dot(K.transpose(K.expand_dims(weight_matrix,axis=-1)),K.expand_dims(K.log(K.flatten(tf.math.reduce_sum(tf.multiply(y_true,y_pre),-1))),axis=-1))
    return E[:,0]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import hashlib
import os

import numpy as np


def calculate_file_hash(file_path, cache_results=False, force_rewrite_cache=False):
    if not cache_results:
        return hashlib.sha256(open(file_path, 'rb').read()).hexdigest()
    
    # check if the hash exists on disk
    hash_file = file_path + '.sha256'
    if not force_rewrite_cache:
        try:
            with open(hash_file, 'r') as f:
                output_hash = f.readline().strip()
        except OSError:
            print('Error while reading from hash file')
        else:
            if len(output_hash) == 64:
                #print('Using cached hash')
                return output_hash
            else:
                print('Invalid stored hash')

    # file does not exist, or stored hash is invalid
    output_hash = calculate_file_hash(file_path, False)  # fallback to calculating hash
    try:
        with open(hash_file, 'w') as f:
            f.write(output_hash)
    except OSError:
        print('Error writing hash to file')
    return output_hash


def calc_dice_score(y_true, y_pred):
    """
    Binary f1. Same results as sklearn f1 binary.

    y_true: 1D / 2D / 3D array
    y_pred: 1D / 2D / 3D array
    """
    intersect = np.sum(y_true * y_pred)  # works because all multiplied by 0 gets 0
    denominator = np.sum(y_true) + np.sum(y_pred)  # works because all multiplied by 0 gets 0
    return (2 * intersect) / (denominator + 1e-6)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
from abc import ABC, abstractmethod
from io import BytesIO
from typing import IO, Callable, List, Union, Optional, Dict
import numpy as np
import time


class IncompatibleModelError(Exception):
    pass

class DeepLearningClass(ABC):   
    
    @abstractmethod
    def init_model(self):
        """
        Initializes the model when needed

        Returns
        -------
        None.

        """
        pass
    
    @abstractmethod
    def calc_delta(self, baseModel: DeepLearningClass) -> DeepLearningClass:
        """
        Calculate a delta with another model, Returns a new instance

        Parameters
        ----------
        baseModel : DeepLearningClass
            Base model to calculate the delta from

        Returns
        -------
        DeepLearningClass
            A deep learning class representing the delta of the two models

        """
        pass
    
    def __sub__(self, rhs):
        return self.calc_delta(rhs)
    
    @abstractmethod
    def apply_delta(self, delta_model: DeepLearningClass) -> DeepLearningClass:
        """
        Applies a delta to this class and returns a new model with the delta applied
        

        Parameters
        ----------
        delta_model : DeepLearningClass
            Applies a delta to the current model

        Returns
        -------
        DeepLearningClass
            The model that is the current model plus the delta

        """
        pass

    def sum(self, other):
        """
        Note: this defaults to apply_delta. Redefine to change behavior
        """
        return self.apply_delta(other)

    def __add__(self, rhs):
        return self.sum(rhs)
    
    @abstractmethod
    def incremental_learn(self, training_data: dict, training_outputs: str):
        """
        Perform an incremental learning step on the given training data/outputs

        Parameters
        ----------
        training_data : Dictionary
            Contains the path to the training data and resolution.
        training_outputs : String
            Contains the path to the training labels.

        Returns
        -------
        None.

        """
        pass
    
    @abstractmethod
    def apply(self, data: dict):
        """
        Applies the deep learning model to the image

        Parameters
        ----------
        data : Dictionary
            Contains the data and the extra information (for example image and resolution)

        Returns
        -------
        Depends on the operation performed:
            For classifiers: str - Containing the label of the image
            For segmenters: dict[str, mask] - Containing the labels and the corresponding 2D masks

        """
        pass

    @abstractmethod
    def factor_multiply(self, factor: float):
        """
        Multiplies all the weights by a float factor
        
        """
        pass

    def __mul__(self, factor: float):
        if not isinstance(factor, (int, float)):
            raise NotImplementedError('Incompatible types for multiplication (only multiplication by numeric factor is allowed)')
        return self.factor_multiply(factor)

    def __rmul__(self, factor: float):
        if not isinstance(factor, (int, float)):
            raise NotImplementedError('Incompatible types for multiplication (only multiplication by numeric factor is allowed)')
        return self.factor_multiply(factor)

    def __call__(self, data: dict):
        return self.apply(data)

    def reset_timestamp(self):
        self.timestamp_id = int(time.time())

class ModelProvider(ABC):
    """
        Abstract class that is the base for loading (and, in the future, storing?) models.
        Has to be subclassed to support local and remote loading.
    """
    
    @abstractmethod
    def load_model(self, model_name: str, progress_callback: Optional[Callable[[int, int], None]] = None,
                   force_download: bool = False,
                   timestamp: Optional[Union[int,str]] = None) -> DeepLearningClass:
        """
        Loads a deep learning model.

        Parameters
        ----------
        model_name : str
            The name of the model to load.
        progress_callback: Callable[[int, int], None] (optional)
            Callback function for progress
        force_download: bool
            Sets the forced redownload of models
        timestamp: int or None
            Return a specific model version (default: latest)

        Returns
        -------
        The model object.

        """
        pass

    @abstractmethod
    def model_details(self, model_name: str) -> dict:
        pass

    @abstractmethod
    def available_models(self) -> Optional[list[str]]:
        """
        Parameters
        ----------
        None

        Returns
        -------
        List of available models
        """
        pass

    @abstractmethod
    def upload_model(self, model_name: str, model: DeepLearningClass, dice_score: float=0.0) -> None:
        """
        Parameters
        ----------
        model_name : str
            The name of the model to upload.
        model:
            The model to be uploaded
        dice_score:
            The average dice score of the client
        """
        pass

    def upload_data(self, data: dict) -> None:
        """
        Uploads data to the server. Converts the data into a stream before calling _upload_bytes

        Parameters
        ----------
        data: dict
            Dictionary of objects that can be saved by Numpy and loaded without using pickle (which is unsafe)
        """
        bytes_io = BytesIO()
        np.savez_compressed(bytes_io, **data)
        self._upload_bytes(bytes_io)
        bytes_io.close()


    @abstractmethod
    def _upload_bytes(self, data: IO):
        """
        Uploads generic data to the server. This is an internal function that implements the server communication.
        The actual function to be called by the client is upload_data with a dict

        Parameters
        ----------
        data: IO
            byte stream that is sent to the server.
        """
        pass

    @abstractmethod
    def log(self, msg: str):
        """
        Sends a message to the server to be logged

        Parameters
        ----------
        msg: str
            the message.
        """
        pass


# -*- coding: utf-8 -*-

#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import SimpleITK as sitk

DEFAULT_BIAS_CORRECTION_LEVELS = 8


def biascorrection(file_or_image, levels=DEFAULT_BIAS_CORRECTION_LEVELS):
    if type(file_or_image) == str:
        return biascorrection_file(file_or_image, levels)
    else:
        return biascorrection_image(file_or_image, levels)


def biascorrection_image(image, levels=DEFAULT_BIAS_CORRECTION_LEVELS):
    MAX_GRAY_VALUE = 600
    if not type(image) == sitk.SimpleITK.Image:
        # normalize values
        image = image*MAX_GRAY_VALUE/image.max()
        image = sitk.GetImageFromArray(image)
        image = sitk.Cast(image, sitk.sitkFloat32)
    else:
        image = sitk.GetArrayFromImage(image)
        image = image * MAX_GRAY_VALUE / image.max()
        image = sitk.GetImageFromArray(image)
        image = sitk.Cast(image, sitk.sitkFloat32)

    maskImage = sitk.OtsuThreshold(image, 0, 1, 200)
    corrector = sitk.N4BiasFieldCorrectionImageFilter()
    numberFittingLevels = levels
    numberOfIteration = [50]
    corrector.SetMaximumNumberOfIterations(numberOfIteration * numberFittingLevels)
    output = corrector.Execute(image, maskImage)
    img2 = sitk.GetArrayFromImage(output)
    return img2


def biascorrection_file(nifti_file, levels=DEFAULT_BIAS_CORRECTION_LEVELS):
    inputImage = sitk.ReadImage(nifti_file,sitk.sitkFloat32) 
    return biascorrection_image(inputImage, levels)



#  Copyright (c) 2021 Dafne-Imaging Team
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-

VERSION='1.3-alpha6'
__version__ = VERSION

from . import resources

import sys
import flexidep

assert sys.version_info.major == 3, "This software is only compatible with Python 3.x"

if sys.version_info.minor < 10:
    import importlib_resources as pkg_resources
else:
    import importlib.resources as pkg_resources

# install the required resources
if not flexidep.is_frozen():
    with pkg_resources.files(resources).joinpath('runtime_dependencies.cfg').open() as f:
        dm = flexidep.DependencyManager(config_file=f)
    dm.install_auto()

from .DynamicDLModel import DynamicDLModel
from .LocalModelProvider import LocalModelProvider
from .RemoteModelProvider import RemoteModelProvider